sig = Signature(3, 1)
cache = ExpressionCache(sig)
sc(x) = scalar(cache, x)
fac(x) = factor(cache, x)
bl(args...) = blade(cache, args...)
x, y, z = sc.([:x, :y, :z])
e1, e2, e3, e4 = blade.(cache, [1, 2, 3, 4])

expression(head, args...) = Expression(cache, head, args...)
add(x, ys...) = expression(SCALAR_ADDITION, x, ys...)
mul(x, ys...) = expression(SCALAR_PRODUCT, x, ys...)
uadd(x, ys...) = unsimplified_expression(cache, SCALAR_ADDITION, x, ys...)
umul(x, ys...) = unsimplified_expression(cache, SCALAR_PRODUCT, x, ys...)

@testset "Factorization" begin
  a, b, c, d, e, f = (:a, :b, :c, :d, :e, :f)
  ex = add(mul(a, b), mul(a, c))
  fact = Factorization(ex)
  @test apply!(fact) ≈ umul(a, uadd(b, c))

  ex = add(mul(a, c), mul(a, d), mul(b, c), mul(b, d), e)
  @test factorize(ex) ≈ uadd(umul(add(a, b), add(c, d)), e)

  ex = add(mul(a, c, e), mul(a, c, f), mul(a, d, e), mul(a, d, f), mul(b, c, e), mul(b, c, f), mul(b, d, e), mul(b, d, f))
  @test factorize(ex) ≈ umul(add(a, b), add(c, d), add(e, f))

  ex2 = fac(ex)
  factorize!(ex2)
  @test factorize(ex) ≈ ex2[1]

  ex = add(mul(a, b, c), mul(c, d, e), mul(a, d, f))
  @test factorize(ex) in (
    uadd(umul(a, add(mul(b, c), mul(d, f))), mul(c, d, e)),
    uadd(umul(c, uadd(mul(a, b), mul(d, e))), mul(a, d, f)),
    uadd(umul(d, add(mul(c, e), mul(a, f))), mul(a, b, c)),
  )

  ex = generate_expression(Signature(3), quote
    Π = a::Vector ⟑ b::Vector
    Ω = exp((-α::Scalar / 2::Scalar) ⟑ Π)
  end; factorize = false, optimize = false)
  factorize!(ex)

  ex = generate_expression(Signature(3), :((x::Vector * x::Bivector ∧ x::Vector + 2::e12)::Multivector); factorize = false, optimize = false)
  @test all(>(1) ∘ length, gather_scalar_expressions(ex))
  factorize!(ex)
  @test all(>(1) ∘ length, gather_scalar_expressions(ex))
end;
